import type { RequestHandler } from '@sveltejs/kit';
import { getSession } from '$lib/server/session';
import { Permissions } from '$lib/permissions';
import { Admin, Report, StrikeReason, User } from '$lib/server/database';
import type { Attributes } from 'sequelize';
import { Op } from 'sequelize';
import { env } from '$env/dynamic/private';
import crypto from 'crypto';
import { webhookUserReported } from '$lib/server/webhook';
import { ReportAddSchema, ReportEditSchema, ReportListSchema } from './schema';

export const POST = (async ({ request, cookies }) => {
	if (getSession(cookies, { permissions: [Permissions.Reports] }) == null) {
		return new Response(null, {
			status: 401
		});
	}

	const parseResult = await ReportListSchema.safeParseAsync(await request.json());
	if (!parseResult.success) {
		return new Response(null, {
			status: 400
		});
	}
	const data = parseResult.data;

	let reportFindOptions: Attributes<Report> = {};
	reportFindOptions.draft = data.draft;
	reportFindOptions.status = data.status ?? ['none', 'review'];
	if (data.reporter != null) {
		const reporter_ids = await User.findAll({
			attributes: ['id'],
			where: { username: { [Op.like]: `%${data.reporter}%` } }
		});
		reportFindOptions.reporter_id = reporter_ids.map((u) => u.id);
	}
	if (data.reported != null) {
		const reported_ids = await User.findAll({
			attributes: ['id'],
			where: { username: { [Op.like]: `%${data.reported}%` } }
		});
		reportFindOptions.reported_id = reported_ids.map((u) => u.id);
	}

	if (data.hash != null) {
		reportFindOptions = { url_hash: data.hash };
		data.from = 0;
		data.limit = 1;
	}

	let reports = await Report.findAll({
		where: reportFindOptions,
		include: [{ all: true }],
		order: [['created_at', 'DESC']],
		offset: data.from || 0,
		limit: data.limit || 100
	});
	reports = reports.map((r) => {
		if (r.auditor_id === null && r.status != 'none') {
			// if the report was edited by the admin account set by the env variable, it has no relation to the admin
			// table in the database, so it gets manually created here. we just assume that the auditor_id is never null
			// when not edited by the env admin account
			r.auditor_id = -1;
			r.dataValues.auditor = {
				id: -1,
				username: env.ADMIN_USER,
				permissions: new Permissions(Permissions.allPermissions()),
				createdAt: 0,
				updatedAt: 0
			};
		} else if (r.auditor) {
			delete r.dataValues.auditor.password;
		}
		if (!r.strike_reason) {
			r.strike_reason_id = -1;
		}
		return r;
	});

	return new Response(
		JSON.stringify({ reports: reports, count: await Report.count({ where: reportFindOptions }) })
	);
}) satisfies RequestHandler;

export const PATCH = (async ({ request, cookies }) => {
	if (getSession(cookies, { permissions: [Permissions.Reports] }) == null) {
		return new Response(null, {
			status: 401
		});
	}

	const parseResult = await ReportEditSchema.safeParseAsync(await request.json());
	if (!parseResult.success) {
		return new Response(null, {
			status: 400
		});
	}
	const data = parseResult.data;

	const report = await Report.findOne({ where: { id: data.id }, include: [{ all: true }] });
	const admin = await Admin.findOne({ where: { id: data.auditor } });
	const reported = data.reported ? await User.findOne({ where: { uuid: data.reported } }) : null;
	if (report === null || (admin === null && data.auditor != -1))
		return new Response(null, { status: 400 });

	const webhookTriggerUsers: string[] = [];

	// check if strike reason has changed and return 400 if it doesn't exist
	if (
		(report.strike_reason?.id ?? -1) != data.strike_reason &&
		data.strike_reason != null &&
		data.strike_reason != -1
	) {
		const strike_reason = await StrikeReason.findByPk(data.strike_reason);
		if (strike_reason == null) return new Response(null, { status: 400 });
	}

	if (data.status === 'reviewed') {
		// trigger webhook if status changed to reviewed
		if (report.status !== 'reviewed' && data.strike_reason != -1 && reported) {
			webhookTriggerUsers.push(reported.uuid!);
		}
		// trigger webhook if strike reason has changed
		else if (
			(report.strike_reason?.id ?? -1) != data.strike_reason &&
			report.reported &&
			reported
		) {
			webhookTriggerUsers.push(reported.uuid!);
		}
	} else if (report.status === 'reviewed') {
		// trigger webhook if report status is reviewed and reported user has changed
		if (report.strike_reason != null && report.reported) {
			webhookTriggerUsers.push(report.reported.uuid!);
		}
	}

	if (data.notice != null) report.notice = data.notice;
	if (data.statement != null) report.statement = data.statement;
	if (data.status != null) report.status = data.status;
	if (data.strike_reason != null)
		report.strike_reason_id = data.strike_reason == -1 ? null : data.strike_reason;
	if (data.strike_reason != null)
		report.striked_at = data.strike_reason == -1 ? null : new Date(Date.now());
	if (admin != null) report.auditor_id = admin.id;

	await report.save();

	for (const webhookTriggerUser of webhookTriggerUsers) {
		// no `await` to avoid blocking
		webhookUserReported(env.REPORTED_WEBHOOK, webhookTriggerUser).catch((e) =>
			console.error(`failed to send reported webhook: ${e}`)
		);
	}

	return new Response();
}) satisfies RequestHandler;

export const PUT = (async ({ request, cookies }) => {
	if (getSession(cookies, { permissions: [Permissions.Reports] }) == null) {
		return new Response(null, {
			status: 401
		});
	}

	const parseResult = await ReportAddSchema.safeParseAsync(await request.json());
	if (!parseResult.success) {
		return new Response(null, {
			status: 400
		});
	}
	const data = parseResult.data;

	const reporter = await User.findOne({ where: { uuid: data.reporter } });
	const reported = data.reported ? await User.findOne({ where: { uuid: data.reported } }) : null;

	if (reporter == null) return new Response(null, { status: 400 });

	const report = await Report.create({
		subject: data.reason,
		body: data.body,
		draft: data.body === null,
		status: 'none',
		url_hash: crypto.randomBytes(18).toString('hex'),
		completed: false,
		reporter_id: reporter.id,
		reported_id: reported?.id || null
	});
	report.dataValues.reporter = await User.findOne({ where: { id: report.reporter_id } });
	report.dataValues.reported = report.reported_id
		? await User.findOne({ where: { id: report.reported_id } })
		: null;
	report.dataValues.auditor = null;

	return new Response(JSON.stringify(report), {
		status: 201
	});
}) satisfies RequestHandler;
