import { getSession } from '$lib/server/session';
import { Permissions } from '$lib/permissions';
import type { RequestHandler } from '@sveltejs/kit';
import {
	ApiError,
	getBedrockUuid,
	getJavaUuid,
	RateLimitError,
	UserNotFoundError
} from '$lib/server/minecraft';
import type { ZodType } from 'zod';
import { FindUuidSchema } from './schema';

export const POST = (async ({ url, request, cookies }) => {
	if (getSession(cookies, { permissions: [Permissions.Tools] }) == null) {
		return new Response(null, { status: 401 });
	}

	const action = url.searchParams.get('action');
	if (!action) {
		return new Response(null, { status: 400 });
	}

	try {
		switch (action) {
			case 'getuuid': {
				const data = await parseData(FindUuidSchema, await request.json());
				return new Response(await findUuid(data.username, data.edition), { status: 200 });
			}
		}
		return new Response(null, { status: 400 });
	} catch (error) {
		return new Response(JSON.stringify({ error: error }), { status: 400 });
	}
}) satisfies RequestHandler;

async function parseData<Output>(schema: ZodType<Output>, json: string): Promise<Output> {
	const parseResult = await schema.safeParseAsync(json);
	if (!parseResult.success) throw new Error(parseResult.error.toString());

	return parseResult.data;
}

async function findUuid(username: string, edition: 'java' | 'bedrock'): Promise<string> {
	let uuid = '';

	try {
		switch (edition) {
			case 'java':
				uuid = await getJavaUuid(username);
				break;
			case 'bedrock':
				uuid = await getBedrockUuid(username);
				break;
		}
	} catch (e) {
		if (e instanceof UserNotFoundError) {
			throw `Der Spielername ${username} existiert nicht`;
		} else if (e instanceof ApiError) {
			throw (e as Error).message;
		} else if (e instanceof RateLimitError) {
			throw 'Rate limit exceeded, bitte versuche es erneut';
		} else {
			throw e;
		}
	}

	return JSON.stringify({
		uuid: uuid
	});
}
